from ajenti.api import *
from ajenti.plugins import *

info = PluginInfo(
    title='Branding',
    icon='asterisk',
    dependencies=[
        PluginDependency('configurator'),
    ]
)

#raise Exception()

# -------------------------------------------------------------------------------------------------
# Hide licensing UI

try:
    import ajenti.licensing
    ajenti.licensing.Licensing.licensing_active = False
except:
    pass


# -------------------------------------------------------------------------------------------------
# Patch the Configure section

from ajenti.plugins.configurator.configurator import Configurator

old_init = Configurator.init


def new_init(self, *args):
    old_init(self, *args)

    # patch the label
    lbl = self.nearest(lambda x: x.typeid == 'label' and x.style == 'small')[0]
    lbl.text = 'Custom text'

    # or just hide it completely
    # self.nearest(lambda x: x.typeid == 'formgroup' and x.text == _('Other'))[0].delete()

Configurator.init = new_init



from __future__ import unicode_literals

import exconsole
import locale
import logging
import os
import signal
import socket
import syslog
import sys


import ajenti
import ajenti.locales  # importing locale before everything else!

import ajenti.feedback
import ajenti.licensing
import ajenti.plugins
from ajenti.http import HttpRoot
from ajenti.middleware import SessionMiddleware, AuthenticationMiddleware
from ajenti.plugins import manager
from ajenti.routing import CentralDispatcher
from ajenti.ui import Inflater

import gevent
import gevent.ssl
from gevent import monkey

import ajenti.ipc

try:
    monkey.patch_all(select=False, thread=False, subprocess=True)
except:
    monkey.patch_all(select=False, thread=False) # old gevent

import ajenti.compat

from socketio.server import SocketIOServer


def run():
    ajenti.init()

    reload(sys)
    sys.setdefaultencoding('utf8')

    try:
        locale.setlocale(locale.LC_ALL, '')
    except:
        logging.warning('Couldn\'t set default locale')

    logging.info('Ajenti %s running on platform: %s' % (ajenti.version, ajenti.platform))

    if ajenti.debug:
        def cmd_list_instances(ctx=None):
            import pprint
            if not ctx:
                from ajenti.plugins import manager
                ctx = manager.context
            pprint.pprint(ctx._get_all_instances())

        def cmd_sessions():
            import pprint
            sessions = SessionMiddleware.get().sessions
            return sessions

        def cmd_list_instances_session():
            cmd_list_instances(cmd_sessions().values()[0].appcontext)

        exconsole.register(commands=[
            ('_manager', 'PluginManager', ajenti.plugins.manager),
            ('_instances', 'return all @plugin instances', cmd_list_instances),
            ('_sessions', 'return all Sessions', cmd_sessions),
            ('_instances_session', 'return all @plugin instances in session #0', cmd_list_instances_session),
        ])

    # Load plugins
    ajenti.plugins.manager.load_all()
    Inflater.get().precache()

    bind_spec = (ajenti.config.tree.http_binding.host, ajenti.config.tree.http_binding.port)
    if ':' in bind_spec[0]:
        addrs = socket.getaddrinfo(bind_spec[0], bind_spec[1], socket.AF_INET6, 0, socket.SOL_TCP)
        bind_spec = addrs[0][-1]

    # Fix stupid socketio bug (it tries to do *args[0][0])
    socket.socket.__getitem__ = lambda x, y: None

    logging.info('Starting server on %s' % (bind_spec, ))
    if bind_spec[0].startswith('/'):
        if os.path.exists(bind_spec[0]):
            os.unlink(bind_spec[0])
        listener = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            listener.bind(bind_spec[0])
        except:
            logging.error('Could not bind to %s' % bind_spec[0])
            sys.exit(1)
        listener.listen(10)
    else:
        listener = socket.socket(socket.AF_INET6 if ':' in bind_spec[0] else socket.AF_INET, socket.SOCK_STREAM)
        try:
            listener.setsockopt(socket.IPPROTO_TCP, socket.TCP_CORK, 1)
        except:
            try:
                socket.TCP_NOPUSH = 4
                listener.setsockopt(socket.IPPROTO_TCP, socket.TCP_NOPUSH, 1)
            except:
                logging.warn('Could not set TCP_CORK/TCP_NOPUSH')
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            listener.bind(bind_spec)
        except:
            logging.error('Could not bind to %s' % (bind_spec,))
            sys.exit(1)
        listener.listen(10)

    stack = [
        SessionMiddleware.get(),
        AuthenticationMiddleware.get(),
        CentralDispatcher.get()
    ]

    ssl_args = {}
    if ajenti.config.tree.ssl.enable:
        ssl_args['certfile'] = ajenti.config.tree.ssl.certificate_path

    ajenti.server = SocketIOServer(
        listener,
        log=open(os.devnull, 'w'),
        application=HttpRoot(stack).dispatch,
        policy_server=False,
        resource='ajenti:socket',
        **ssl_args
    )

    # auth.log
    try:
        syslog.openlog(
            ident=str(b'ajenti'),
            facility=syslog.LOG_AUTH,
        )
    except:
        syslog.openlog(b'ajenti')

    try:
        gevent.signal(signal.SIGINT, lambda: sys.exit(0))
        gevent.signal(signal.SIGTERM, lambda: sys.exit(0))
    except:
        pass

    ajenti.feedback.start()
    ajenti.ipc.IPCServer.get().start()
    ajenti.licensing.Licensing.get()

    ajenti.server.serve_forever()

    if hasattr(ajenti.server, 'restart_marker'):
        logging.warn('Restarting by request')

        fd = 20  # Close all descriptors. Creepy thing
        while fd > 2:
            try:
                os.close(fd)
                logging.debug('Closed descriptor #%i' % fd)
            except:
                pass
            fd -= 1

        os.execv(sys.argv[0], sys.argv)
    else:
        logging.info('Stopped by request')



from ajenti.api import *
from ajenti.plugins import *


info = PluginInfo(
    title='File Manager',
    icon='folder-open',
    dependencies=[
        PluginDependency('main'),
        PluginDependency('tasks'),
    ],
)


def init():
    import fm



# (c) 2007 Chris AtLee <chris@atlee.ca>
# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license.php
"""
PAM module for python

Provides an authenticate function that will allow the caller to authenticate
a user against the Pluggable Authentication Modules (PAM) on the system.

Implemented using ctypes, so no compilation is necessary.
"""
__all__ = ['authenticate']

from ctypes import CDLL, POINTER, Structure, CFUNCTYPE, cast, pointer, sizeof
from ctypes import c_void_p, c_uint, c_char_p, c_char, c_int
from ctypes.util import find_library

LIBPAM = CDLL(find_library("pam"))
LIBC = CDLL(find_library("c"))

CALLOC = LIBC.calloc
CALLOC.restype = c_void_p
CALLOC.argtypes = [c_uint, c_uint]

STRDUP = LIBC.strdup
STRDUP.argstypes = [c_char_p]
STRDUP.restype = POINTER(c_char) # NOT c_char_p !!!!

# Various constants
PAM_PROMPT_ECHO_OFF = 1
PAM_PROMPT_ECHO_ON = 2
PAM_ERROR_MSG = 3
PAM_TEXT_INFO = 4

class PamHandle(Structure):
    """wrapper class for pam_handle_t"""
    _fields_ = [
            ("handle", c_void_p)
            ]

    def __init__(self):
        Structure.__init__(self)
        self.handle = 0

class PamMessage(Structure):
    """wrapper class for pam_message structure"""
    _fields_ = [
            ("msg_style", c_int),
            ("msg", POINTER(c_char)),
            ]

    def __repr__(self):
        return "<PamMessage %i '%s'>" % (self.msg_style, self.msg)

class PamResponse(Structure):
    """wrapper class for pam_response structure"""
    _fields_ = [
            ("resp", POINTER(c_char)),
            ("resp_retcode", c_int),
            ]

    def __repr__(self):
        return "<PamResponse %i '%s'>" % (self.resp_retcode, self.resp)

CONV_FUNC = CFUNCTYPE(c_int,
        c_int, POINTER(POINTER(PamMessage)),
               POINTER(POINTER(PamResponse)), c_void_p)

class PamConv(Structure):
    """wrapper class for pam_conv structure"""
    _fields_ = [
            ("conv", CONV_FUNC),
            ("appdata_ptr", c_void_p)
            ]

PAM_START = LIBPAM.pam_start
PAM_START.restype = c_int
PAM_START.argtypes = [c_char_p, c_char_p, POINTER(PamConv),
        POINTER(PamHandle)]

PAM_END = LIBPAM.pam_end
PAM_END.restpe = c_int
PAM_END.argtypes = [PamHandle, c_int]

PAM_AUTHENTICATE = LIBPAM.pam_authenticate
PAM_AUTHENTICATE.restype = c_int
PAM_AUTHENTICATE.argtypes = [PamHandle, c_int]

def authenticate(username, password, service='login'):
    """Returns True if the given username and password authenticate for the
    given service.  Returns False otherwise
    
    ``username``: the username to authenticate
    
    ``password``: the password in plain text
    
    ``service``: the PAM service to authenticate against.
                 Defaults to 'login'"""
    @CONV_FUNC
    def my_conv(n_messages, messages, p_response, app_data):
        """Simple conversation function that responds to any
        prompt where the echo is off with the supplied password"""
        # Create an array of n_messages response objects
        addr = CALLOC(n_messages, sizeof(PamResponse))
        p_response[0] = cast(addr, POINTER(PamResponse))
        for i in range(n_messages):
            if messages[i].contents.msg_style == PAM_PROMPT_ECHO_OFF:
                pw_copy = STRDUP(str(password))
                p_response.contents[i].resp = pw_copy
                p_response.contents[i].resp_retcode = 0
        return 0

    handle = PamHandle()
    conv = PamConv(my_conv, 0)
    retval = PAM_START(service, username, pointer(conv), pointer(handle))

    if retval != 0:
        # TODO: This is not an authentication error, something
        # has gone wrong starting up PAM
        PAM_END(handle, retval)
        return False

    retval = PAM_AUTHENTICATE(handle, 0)
    e = PAM_END(handle, retval)
    return retval == 0 and e == 0



from base64 import b64encode
import catcher
import gevent
import gevent.coros
import json
import logging
import requests
import traceback
import zlib

import ajenti
from ajenti.api import *
from ajenti.api.http import *
from ajenti.api.sensors import Sensor
from ajenti.licensing import Licensing
from ajenti.middleware import AuthenticationMiddleware
from ajenti.plugins import manager
from ajenti.profiler import *
from ajenti.users import PermissionProvider, UserManager, SecurityError
from ajenti.ui import *
from ajenti.util import make_report
import ajenti.feedback

from api import SectionPlugin


@plugin
class MainServer (BasePlugin, HttpPlugin):
    @url('/')
    def handle_index(self, context):
        context.add_header('Content-Type', 'text/html')
        if context.session.identity is None:
            context.respond_ok()
            html = self.open_content('static/auth.html').read()
            return html % {
                'license': json.dumps(Licensing.get().get_license_status()),
                'error': json.dumps(context.session.data.pop('login-error', None)),
            }
        context.respond_ok()
        return self.open_content('static/index.html').read()

    @url('/ajenti:auth')
    def handle_auth(self, context):
        username = context.query.getvalue('username', '')
        password = context.query.getvalue('password', '')
        if not AuthenticationMiddleware.get().try_login(
            context, username, password
        ):
            context.session.data['login-error'] = _('Invalid login or password')
            gevent.sleep(3)
        return context.redirect('/')

    @url('/ajenti:auth-persona')
    def handle_persona_auth(self, context):
        assertion = context.query.getvalue('assertion', None)
        audience = context.query.getvalue('audience', None)
        if not assertion:
            return self.context.respond_forbidden()

        data = {
            'assertion': assertion,
            'audience': audience,
        }

        resp = requests.post('https://verifier.login.persona.org/verify', data=data, verify=True)

        if resp.ok:
            verification_data = json.loads(resp.content)
            if verification_data['status'] == 'okay':
                context.respond_ok()
                email = verification_data['email']
                for user in ajenti.config.tree.users.values():
                    if user.email == email:
                        AuthenticationMiddleware.get().login(context, user.name)
                        break
                else:
                    context.session.data['login-error'] = _('Email "%s" is not associated with any user') % email
                return ''
        context.session.data['login-error'] = _('Login failed')
        return context.respond_not_found()

    @url('/ajenti:logout')
    def handle_logout(self, context):
        AuthenticationMiddleware.get().logout(context)
        context.session.destroy()
        return context.redirect('/')


@plugin
class MainSocket (SocketPlugin):
    name = '/stream'
    ui = None

    def on_connect(self):
        self.compression = 'zlib'
        self.__updater_lock = gevent.coros.RLock()

        # Inject into session
        self.request.session.endpoint = self

        # Inject the context methods
        self.context.notify = self.send_notify
        self.context.launch = self.launch
        self.context.endpoint = self

        if not 'ui' in self.request.session.data:
            # This is a newly created session
            ui = UI.new()

            self.request.session.data['ui'] = ui
            ui.root = MainPage.new(ui)
            ui.root.username = self.request.session.identity
            sections_root = SectionsRoot.new(ui)

            for e in sections_root.startup_crashes:
                self.send_crash(e)

            ui.root.append(sections_root)
            ui._sections = sections_root.children
            ui._sections_root = sections_root

        self.ui = self.request.session.data['ui']
        self.send_init()
        self.send_ui()
        self.spawn(self.ui_watcher)

    def on_message(self, message):
        if not self.ui:
            return
        self.spawn(self.handle_message, message)

    def handle_message(self, message):
        try:
            with self.__updater_lock:
                if message['type'] == 'ui_update':
                    # UI updates arrived
                    profile_start('Total')
                    # handle content updates first, before events affect UI
                    for update in message['content']:
                        if update['type'] == 'update':
                            # Property change
                            profile_start('Handling updates')
                            els = self.ui.root.nearest(
                                lambda x: x.uid == update['uid'],
                                exclude=lambda x: x.parent and not x.parent.visible,
                                descend=False,
                            )
                            if len(els) == 0:
                                continue
                            el = els[0]
                            for k, v in update['properties'].iteritems():
                                setattr(el, k, v)
                            profile_end('Handling updates')
                    for update in message['content']:
                        if update['type'] == 'event':
                            # Element event emitted
                            profile_start('Handling event')
                            self.ui.dispatch_event(update['uid'], update['event'], update['params'])
                            profile_end('Handling event')
                    if self.ui.has_updates():
                        # If any updates happened due to event handlers, send these immediately
                        self.ui.clear_updates()
                        self.send_ui()
                    else:
                        # Otherwise just ACK
                        self.send_ack()
                    profile_end('Total')
                    if ajenti.debug:
                        self.send_debug()
        except SecurityError, e:
            self.send_security_error()
        except Exception, e:
            catcher.backup(e)
            traceback.print_exc()
            e.traceback = traceback.format_exc(e)
            self.send_crash(e)

    def send_init(self):
        data = {
            'version': ajenti.version,
            'platform': ajenti.platform,
            'hostname': Sensor.find('hostname').value(),
            'session': self.context.session.id,
            'feedback': ajenti.config.tree.enable_feedback,
            'edition': ajenti.edition,
            'compression': self.compression,
        }
        self.emit('init', json.dumps(data))

    def send_ui(self):
        profile_start('Rendering')
        data = json.dumps(self.ui.render())
        profile_end('Rendering')
        if self.compression == 'zlib':
            data = b64encode(zlib.compress(data, 4)[2:-4])
        if self.compression == 'lzw':
            data = b64encode(self.__compress_lzw(data))
        self.emit('ui', data)

    def __compress_lzw(self, data):
        dict = {}
        out = []
        phrase = data[0]
        code = 256
        for i in range(1, len(data)):
            currChar = data[i]
            if dict.get(phrase + currChar, None):
                phrase += currChar
            else:
                out.append(dict[phrase] if (len(phrase) > 1) else ord(phrase[0]))
                dict[phrase + currChar] = code
                code += 1
                phrase = currChar
        out.append(dict[phrase] if (len(phrase) > 1) else ord(phrase[0]))
        res = ''
        for code in out:
            if code >= 256:
                res += chr(0)
                res += chr(code / 256)
            res += chr(code % 256)
        return res

    def send_ack(self):
        self.emit('ack')

    def send_progress(self, msg=''):
        self.emit('progress-message', msg)
        gevent.sleep(0)

    def send_update_request(self):
        self.emit('update-request')

    def send_security_error(self):
        self.emit('security-error', '')

    def send_open_tab(self, url, title='new tab'):
        self.emit('openTab', json.dumps({'url': url, 'title': title}))

    def send_close_tab(self, url):
        self.emit('closeTab', json.dumps({'url': url}))

    def send_debug(self):
        profiles = get_profiles()
        logging.debug(repr(profiles))
        data = {
            'profiles': profiles
        }
        self.emit('debug', json.dumps(data))

    def send_crash(self, exc):
        if not hasattr(exc, 'traceback'):
            traceback.print_exc()
            exc.traceback = traceback.format_exc(exc)
        data = {
            'message': str(exc),
            'traceback': exc.traceback,
            'report': make_report(exc)
        }
        data = json.dumps(data)
        self.emit('crash', data)

    def send_notify(self, type, text):
        self.emit('notify', json.dumps({'type': type, 'text': text}))

    def switch_tab(self, tab):
        self.ui._sections_root.on_switch(tab.uid)

    def launch(self, intent, **data):
        for section in self.ui._sections:
            for handler in section._intents:
                if handler == intent:
                    section._intents[handler](**data)
                    return

    def get_section(self, cls):
        for section in self.ui._sections:
            if section.__class__ == cls:
                return section

    def ui_watcher(self):
        # Sends UI updates periodically
        while True:
            if self.__updater_lock:
                if self.ui.has_updates():
                    self.send_update_request()
                    gevent.sleep(0.5)
            gevent.sleep(0.2)


@plugin
class SectionPermissions (PermissionProvider):
    def get_name(self):
        return _('Section access')

    def get_permissions(self):
        # Generate permission set on-the-fly
        return sorted([
            ('section:%s' % x.__class__.__name__, (_(x.category) or 'Ajenti') + ' | ' + _(x.title))
            for x in SectionPlugin.get_instances()
        ], key=lambda x: x[1])


@p('username')
@plugin
class MainPage (UIElement, BasePlugin):
    typeid = 'main:page'


@p('name')
@plugin
class SectionsCategory (UIElement):
    typeid = 'main:sections_category'


@p('is_empty', type=bool)
@plugin
class SectionsRoot (UIElement):
    typeid = 'main:sections_root'
    category_order = {
        '': 0,
        'System': 50,
        'Web': 60,
        'Tools': 70,
        'Software': 80,
        'Other': 99
    }

    def init(self):
        self.categories = {}
        self.startup_crashes = []

        profile_start('Starting plugins')

        self.is_empty = True
        for cls in SectionPlugin.get_classes():
            try:
                UserManager.get().require_permission(self.context, 'section:%s' % cls.__name__)

                try:
                    profile_start('Starting %s' % cls.__name__)
                    cat = cls.new(self.ui)
                    cat.clsname = cls.classname
                    profile_end()
                    self.append(cat)
                    self.is_empty = False
                except SecurityError:
                    pass
                except Exception, e:
                    catcher.backup(e)
                    traceback.print_exc()
                    e.traceback = traceback.format_exc(e)
                    self.startup_crashes.append(e)
            except SecurityError:
                pass

        profile_end()

        def category_order(x):
            order = 98
            if x.category in self.category_order:
                order = self.category_order[x.category]
            return (order, x.order, x.title)

        self.children = sorted(self.children, key=category_order)
        if len(self.children) > 0:
            self.on_switch(self.children[0].uid)

    def on_switch(self, uid):
        for child in self.children:
            child.active = child.uid == uid
            child.visible = child.active
            if child.active:
                if child._first_page_load:
                    child.broadcast('on_first_page_load')
                    child._first_page_load = False
                child.broadcast('on_page_load')
        self.invalidate()
